# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# fmt: off
from ......schema_classes import DatahubIngestionCheckpointClass
from ......schema_classes import DatahubIngestionRunSummaryClass
from ......schema_classes import IngestionCheckpointStateClass


DatahubIngestionCheckpoint = DatahubIngestionCheckpointClass
DatahubIngestionRunSummary = DatahubIngestionRunSummaryClass
IngestionCheckpointState = IngestionCheckpointStateClass

# fmt: on
